/* Reminder to PHil: dont forget sun sample driver pvip! */


/* This is a quickie prototype for a sparc-style /dev/fb type driver,
 * for use with Solaris x86
 *
 * It was thrown together using my "pci base" template, from
 *  http://www.bolthole.com/solaris/drivers/
 *
 *
 * Read the README file for important information
 *
 *    fb.c version 1.6
 * 
 * You can use this source for any purpose you like [without any warantee, etc]
 * It would be nice if you give me some credit if you use it, but you dont
 * have to.
 *
 * REMEMBER: If a C function isn't mentioned in
 *  "man -s9f intro", you pretty much cannot use it here.
 *
 *
 * DISCLAIMER: This code may not be 100% clean, kernel-wise.
 *  But it is as clean as I know how to make it, with lots of
 *  hopefully helpful comments.
 *    Philip Brown
 */

#include <varargs.h>
#define _SYS_VARARGS_H		/* sys/varargs.h hack for gcc */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/pci.h>
#include <sys/fbio.h>

/* THESE MUST BE LAST */
#include <sys/ddi.h>
#include <sys/sunddi.h>



/* This ASSUMES you create minor nodes where the higher bits denote
 * the instance of the driver. (aka "which card is this?")
 * The lower bits denote clone number, or anything else you want.
 */

#define DEVT_TO_INST(devt) (getminor(devt)>>4)
#define DEVT_TO_CLONENUM(devt) (getminor(devt) & 0xf)

/* Used to choose a node  num for a device node compatible with DEVT_TO_INST*/
#define INST_TO_NODENUM(instance) (instance<<4)


#define DEVNODE_NAME	"fb"	
#define DEVNODE_TYPE	DDI_NT_DISPLAY /* see sunddi.h */

/*Our globally visible stuffs, that we populate the global
 * driver entry structs with  [base_cb_ops, base_ops]
 * Yeah, its kind of wierd to have a "static" be also externally visible,
 * but that's traditionally how these are declared.
 */
static	int	fb_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	fb_probe(dev_info_t *dip);
static	int	fb_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	fb_open(dev_t *dev, int openflags, int otyp, cred_t *credp);

static	int	fb_print(dev_t, char*);

static	int	fb_read(dev_t dev, struct uio* uiop, cred_t * credp);

static	int	fb_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	fb_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	fb_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
			   cred_t *credp, int *rvalp);
static	int	fb_mmap(dev_t dev, off_t, int /* prot */);
static	int	fb_devmap(dev_t dev, devmap_cookie_t, offset_t, size_t,
			  size_t*, uint_t);

static	int	fb_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);


static int debug_level;

/* "char/block operations" */
static struct cb_ops	fb_cb_ops = {
	fb_open,		/* open() */
	fb_close,		/* close() */
	nodev,			/* strategy()		*/
	fb_print,		/* print routine	*/
	nodev,			/* no dump routine	*/
	fb_read,		/* read() */
	fb_write,		/* write() */
	fb_ioctl,		/* ioctl() */
	fb_devmap,		/* devmap routine	*/
	nodev,		/* mmap() */
	ddi_devmap_segmap,	/* default segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP | D_DEVMAP,	/* safe for multi-thread/multi-processor */
	CB_REV,			/* cb_ops version? */
	nodev,			/* cb_aread() */
	nodev,			/* cb_awrite() */
};

/* "device operations". */
static struct dev_ops fb_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	fb_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	fb_probe,		/* devo_probe */
	fb_attach,		/* devo_attach */
	fb_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&fb_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

/* Ignore this. This is only a filler for the next struct, and
 * is already declared in <sys/modctl.h>
 */
extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"Experimental _fb_ driver",
	&fb_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 is indicated by manual */
	{&modldrv,NULL,NULL,NULL}

};


/************************************************************
 * And here's finally the vars specific to THIS driver,
 * that get allocated on a per-instance basis.
 * Here's where you should keep instance-specific mutexes,
 * DMA handles, register mapping handles, and stuff like that
 *
 ************************************************************/


/* This describes things like byteorder internally for the pci card */
static struct ddi_device_acc_attr dev_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};


#define MAX_INSTANCES 4 /* max of 4 hardware devices loaded */

/* 
 * Note that this mutex gets used in _init. Which means this mutex
 * isn't just for this "instance" of the driver. It is for ALL INSTANCES.
 * If you have multiple cards, you have multiple instances of the driver,
 * with this single shared mutex across them all
 */
static kmutex_t   fb_global_mutex;

/* This is SORT OF like	int (*array)[] , except you NEVER USE THIS DIRECTLY.
 * You just use it with ddi_get_soft_state().
 * 
 *    "Global Softstate Handle" 
 */
void * fb_glob_soft_handle;

/* This gets filled out on a per-instance basis by fb_attach */
struct fb_softstate {
	dev_info_t *dip; /* you'll probably want this saved! */
	kmutex_t instance_mutex;
	kmutex_t intr_mutex;
	ddi_iblock_cookie_t intr_cookie;

	int vmem_regindex;
	size_t vmemsize;
	paddr_t fb_vmemaddr; /* physmem address */

	/* These are for ioctl info */
	struct fbtype mytype;
	struct fbgattr mygattr;
};


/*****************************************************************/
/*****************************************************************/
/*              Start of actual kernel routines here             */
/*****************************************************************/
/*****************************************************************/


void
fb_debugprint(va_alist)
va_dcl
{
	va_list args;
	char *fmt;
	if(debug_level==0){
		return;
	}

	va_start(args);
	fmt = va_arg(args, char *);
	/* DO NOT USE cmn_err here!!! it will not be interrupt-safe */
	vprintf(fmt, args);
	va_end(args);
}



/*Like the fb_global_mutex above, this routine is GLOBAL across all
 * instances.
 * It actually gets called when you add_drv your driver from the command
 * line, before  * it is hooked up to any specific hardware.
 * See fb_attach() below for instance-specific stuff.
 *
 * HOWEVER... Its often nice to pre-allocate "softstate" space
 * here, if you dont mind hardcoding the total number of possible
 * instances you can have. So we do that.
 * 
 */
int
_init(void)
{
       int     ret;

	fb_debugprint("fb: _init called\n");
	/*
               * Initialize global mutex before mod_install'ing driver.
               * If mod_install() fails, must clean up mutex initialization
               */
	mutex_init(&fb_global_mutex, "fb: Global Mutex",
                     MUTEX_DRIVER, (void *)NULL);

	/* This is kind of like
	 * malloc(sizeof(struct fb_softstate *MAX_INSTANCES))
	 */
	ret=ddi_soft_state_init(&fb_glob_soft_handle,
				sizeof(struct fb_softstate),
			        MAX_INSTANCES);
	if(ret!=0){
		return (ret);
	}
	
	if ((ret = mod_install(&modlinkage)) != 0) {
		mutex_destroy(&fb_global_mutex);
		ddi_soft_state_fini(&fb_glob_soft_handle);
		cmn_err(CE_NOTE, "fb: _init: mod_install FAILED:\n");
		return DDI_FAILURE;
	}

	cmn_err(CE_NOTE, "fb: _init returning SUCCESS, supposedly\n");
	return (DDI_SUCCESS);
}

int
_info(struct modinfo *modinfop)
{
	
	cmn_err(CE_NOTE, "fb: _info called\n");
	return (mod_info(&modlinkage, modinfop));
}

/* This is the match to the _init() routine. Clean up stuff allocated
 * specifically in _init() here.
 * Stuff in attach() should be cleaned up in detach()
 */
int
_fini(void)
{
	int ret;

	cmn_err(CE_NOTE, "fb: _fini called\n");
	
        /*
         * If mod_remove() is successful, we destroy our global mutex
         */
	if ((ret = mod_remove(&modlinkage)) == 0) {
		ddi_soft_state_fini(&fb_glob_soft_handle);
		mutex_destroy(&fb_global_mutex);
	}
	return (ret);
}

/************************************************************/

/*
 * interrupt handler for the base device.
 * This wont ever get called until ddi_add_intr has been called.
 * Dont forget that you can only use intr_mutex in here, NOT
 * instance_mutex.
 * Additionally, you can ONLY USE FUNCTIONS that are
 * INTERRUPT CONTEXT SAFE.
 *  "user", or "kernel" context, just wont cut it.
 *
 * Note that cmn_err() is NOT interrupt-context safe.
 */
static uint_t
fb_intr(caddr_t intr_arg)
{
	/* struct fb_softstate *softstate;
	 * softstate = (struct fb_softstate *)intr_arg;
	 */

	/* Put whatever you want here, IF you have interrupts */
	/* dont forget to call ddi_add_intr() to enable it! */

	return DDI_INTR_UNCLAIMED;
}

/* Kinda wierd, but this is an official solaris entry point, that can
 * be registered by cb_ops. We dont use it, but we provide it
 * for the OS
 */
static int fb_print(dev_t dev, char*string){
	int instance=DEVT_TO_INST(dev);
	cmn_err(CE_NOTE, "base-print(%d): %s",instance,string);
	return 0;
}

static int fb_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd,
			     void *arg, void **result){
	fb_debugprint("fb: getinfo called\n");
	return DDI_FAILURE;
}

/* replacement for mmap() driver entrypoint*/
/* Cribbed from from "Writing Device Drivers", docs.sun.com */
static	int fb_devmap(dev_t dev, devmap_cookie_t cookie, offset_t offset,
		      size_t len, size_t *mappedlen, uint_t model)
{
	int instance=DEVT_TO_INST(dev);
	int status;
	struct fb_softstate *softstate;
	softstate=ddi_get_soft_state(fb_glob_soft_handle, instance);
	if(softstate==NULL){
		return (ENXIO);
	}
	len=ptob(btopr(len)); /* make len an even pagesize */
	if(offset + len  >softstate->vmemsize){
		return (-1);
	}

	status=devmap_devmem_setup(cookie, softstate->dip, NULL,
	          softstate->vmem_regindex, offset, len,
		  PROT_ALL, DEVMAP_DEFAULTS,&dev_attr);

	*mappedlen=len;
	/* 'model' is either DDI_MODEL_ILP32 or DDI_MODEL_ILP64 */
	return(status);

}

/* This has been obsoleted by fb_devmap () */
static int fb_mmap(dev_t dev, off_t offset, int prot)
{
	return(-1);
}


/* you probably wont even need to have a probe for a PCI driver 
 * But just in case you want one to tell you what major and minor
 * you're being probed with, here's one
 */
 
static int fb_probe(dev_info_t *dip){
#ifndef DO_PROBE_ROUTINE
   	fb_debugprint("fb: probe called\n");
#else

  	target=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"target", -1);
  	lun=ddi_prop_get_int(DDI_DEV_T_ANY,dip,DDI_PROP_DONTPASS,
				"lun", -1);
   
   	cmn_err(CE_NOTE, "probe- called fortarget=%d, lun=%d\n", target,lun);

	/* Normally, we would check the dev_info info,
	 * to see which device ID is being probed, and then
	 * check to see if the durn thing is actually there.
	 * But for now, assume .conf file is correct
	 */
   
	/* our two options for success are:
	 * DDI_PROBE_DONTCARE
	 * DDI_PROBE_SUCCESS
	 */

#endif /* DO_PROBE_ROUTINE */

	return DDI_PROBE_DONTCARE;
	/*return DDI_PROBE_FAILURE;*/
}

/* This routine can be called by fb_attach (if errors occur)
 * and fb_detach (on normal detach)
 * Idea is to clean up any mutexes, register mappings, etc. nicely.
 * We also free the instance-softstate.
 * We do NOT remove the minor node, however. Only detach does
 * that currently.
 */
static void fb_cleanup(struct fb_softstate *softstate){
	int instance;

	if(softstate==NULL){
		/* yes, paranoia is good*/
		cmn_err(CE_WARN,"ERROR: fb_cleanup called with NULL");
		return;
	}
   	instance=ddi_get_instance(softstate->dip);

#ifdef BIND_INTERRUPT
	if(softstate->intr_cookie != NULL){
		ddi_remove_intr(softstate->dip,0,softstate->intr_cookie);
	}
#endif

	ddi_soft_state_free(fb_glob_soft_handle,instance);

}


/*
 * _init() has already been called. This routine gets called
 * once for EVERY MATCHING PIECE OF HARDWARE.
 * It gets called for initial "attach", and also for resume(/suspend)
 * You would usually allocate space for state variables, and
 * map device registers in here.
 *
 * To avoid reboot loops, be very very careful putting things in here.
 * If you want to put extra debug type stuff in, it's safer putting
 * it in fb_open() or someplace like that, that gets called after
 * the driver is safely loaded
 *
 * Be sure to free/deallocate anything you create here, in
 *  fb_detach()
 */
static int fb_attach(dev_info_t *dip, ddi_attach_cmd_t cmd){
   	int instance;
	struct fb_softstate *softstate;
	int tmpint;

	/* default debug level to **1**
	 * You can override in the .conf file with "debug-level=0".
	 * When you have a production-ready driver, change
	 * default to 0
	 */
	debug_level=ddi_prop_get_int(DDI_DEV_T_ANY,dip, DDI_PROP_DONTPASS,
                                    "debug-level", 0);

	fb_debugprint("fb: fb_attach called\n");

   	switch(cmd){
	 case DDI_RESUME:
	   	cmn_err(CE_NOTE, "fb: Ignoring attach_RESUME\n");
		return DDI_FAILURE;
	 case DDI_PM_RESUME:
		cmn_err(CE_NOTE, "fb: Ignoring attach_PM_RESUME\n");
		return DDI_FAILURE;
	 case DDI_ATTACH:
	   	break;
	 default:
	   	cmn_err(CE_NOTE, "fb: unknown attach command\n");
	   	return DDI_FAILURE;
	}
   	instance=ddi_get_instance(dip);
   	fb_debugprint("fb: attach for instance %d\n", instance);


	/* We dont have to check for being over MAX_INSTANCES, because
	 * these two calls do it implicitly
	 */
	if (ddi_soft_state_zalloc(fb_glob_soft_handle, instance) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "fb: %d: Failed to alloc soft state",
		    instance);
		return (DDI_FAILURE);
	}
	softstate=ddi_get_soft_state(fb_glob_soft_handle,instance);
	if(softstate==NULL){
		cmn_err(CE_WARN, "fb: attach for instance %d: Failed to get soft state",
			instance);
		ddi_soft_state_free(fb_glob_soft_handle,instance);
		return (DDI_FAILURE);
	}

	softstate->dip = dip;

	/* you will probably want this at some point or another */
	mutex_init(&softstate->instance_mutex, NULL, MUTEX_DRIVER, NULL);



	if(ddi_dev_nregs(dip, &tmpint)==DDI_FAILURE){
		tmpint=0;
	}
	/* IFF there are registers to map, map the first set */
	if(tmpint>0){
		pci_regspec_t *regs;
		int length,regcount;
		int vmem_regnum=-1;

		fb_debugprint("fb: %d register sets exist.\n",tmpint);

		/* This is ugly. But there isnt a nicer ddi way of
		 * parsing the registers. sigh
		 */
		
		if(ddi_getlongprop(DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS,
		   "reg", (caddr_t)&regs, &length) != DDI_PROP_SUCCESS) {
			fb_cleanup(softstate);
			return DDI_FAILURE;
		}
		/* yes, 0 in regs is throwaway, 0 in assignedaddr is not */
		for(regcount=1; regcount<tmpint; regcount++){
			fb_debugprint("regset #%d : IO=%s, reloc=%s, pf=%s,addr=%x, len=%x\n",
			       regcount,
	       ((regs[regcount].pci_phys_hi & PCI_ADDR_IO) >0 ? "yes" : "no"),
	       ((regs[regcount].pci_phys_hi & PCI_RELOCAT_B)>0 ? "yes" : "no"),
	       ((regs[regcount].pci_phys_hi & PCI_REG_PF_M)>0 ? "yes" : "no"),
			       regs[regcount].pci_phys_low,
			       regs[regcount].pci_size_low);

			if(((regs[regcount].pci_phys_hi & PCI_ADDR_IO) ==0) &&
			   ((regs[regcount].pci_phys_hi & PCI_RELOCAT_B) ==0) &&
			   ((regs[regcount].pci_phys_hi & PCI_REG_PF_M) !=0))
			{
				vmem_regnum=regcount;
				softstate->vmemsize=
				softstate->mytype.fb_size=
				       regs[regcount].pci_size_low;
				fb_debugprint("  I like this one! using %d, size %x\n",
					      vmem_regnum,softstate->vmemsize);
			}
		}
		if(ddi_getlongprop(DDI_DEV_T_ANY, dip, DDI_PROP_DONTPASS,
		   "assigned-addresses", (caddr_t)&regs, &length) != DDI_PROP_SUCCESS) {
			fb_cleanup(softstate);
			return DDI_FAILURE;
		}
		for(regcount=0; regcount<tmpint-1; regcount++){
			fb_debugprint("addrset #%d : IO=%s, reloc=%s, pf=%s,addr=%x, len=%x\n",
			       regcount+1,
	       ((regs[regcount].pci_phys_hi & PCI_ADDR_IO) >0 ? "yes" : "no"),
	       ((regs[regcount].pci_phys_hi & PCI_RELOCAT_B)>0 ? "yes" : "no"),
	       ((regs[regcount].pci_phys_hi & PCI_REG_PF_M)>0 ? "yes" : "no"),
			       regs[regcount].pci_phys_low,
			       regs[regcount].pci_size_low);

			       
		}

		if(vmem_regnum==-1){
			fb_debugprint("fb: ERR: could not find videomem\n");
			fb_cleanup(softstate);
			return DDI_FAILURE;
		}
		softstate->vmem_regindex=vmem_regnum;
		ddi_dev_regsize(dip,vmem_regnum, &softstate->vmemsize);
		fb_debugprint("according to dev_regsize, size=%x\n",
			      softstate->vmemsize);
		softstate->mytype.fb_size=softstate->vmemsize;

		
	} else {
		fb_debugprint("fb: ERR: there are no registers to map\n");
		fb_cleanup(softstate);
		return DDI_FAILURE;
	}




	/* These are default, currently faked, values */
	softstate->mytype.fb_type= FBTYPE_MEMCOLOR;
	softstate->mytype.fb_height= 768;
	softstate->mytype.fb_width= 1024;
	softstate->mytype.fb_depth= 1024;
	softstate->mytype.fb_cmsize= 0;

	/* This will create a device node in the appropriate place
	 * under /devices/pci[*]/display@??:fb
	 *
	 * It seems devfsadm takes care of making a link under /dev/fbs,
	 * because we advertise this node as type
	 * (DEVNODE_TYPE == DDI_NT_DISPLAY )
	 */

	{
		char nodename[20];
		sprintf(nodename,"%s%d",DEVNODE_NAME,instance);

		ddi_create_minor_node(dip, nodename, S_IFCHR,
			  INST_TO_NODENUM(instance),
			  DEVNODE_TYPE, 0);
	}

#ifdef BIND_INTERRUPT
	/* We want to do interrupt registering late in the game.
	 * Make sure everything is set up fully,
	 * just in case a random interrupt triggers the handler before
	 * we've even fully finished attach() !
	 */
	if(ddi_dev_nintrs(dip, &tmpint) == DDI_FAILURE){
		tmpint=0;
	}
	if(tmpint>0){
		fb_debugprint("fb: interrupt available. Registering.\n");
		/* And if you have interrupts, you should use THIS mutex for
		 * anything that the interrpt handler waits on
		 */
		ddi_get_iblock_cookie(dip, 0, &softstate->intr_cookie);
		mutex_init(&softstate->intr_mutex, NULL,
			   MUTEX_DRIVER, (void*)softstate->intr_cookie);

		if (ddi_add_intr(dip, 0,
		     &softstate->intr_cookie, NULL,
		     fb_intr, (caddr_t)softstate) != DDI_SUCCESS) {

			cmn_err(CE_WARN, "Could not add interrupt handler");
			fb_cleanup(softstate);
			ddi_remove_minor_node(dip, DEVNODE_NAME);
			return DDI_FAILURE;
		}
	}
#endif

	/* Dont forget to make ddi_remove_minor_node match in detach() */

	return DDI_SUCCESS;

	/* if you add anything to this routine, make ultra-paranoid
	 * checks on return values and return (DDI_FAILURE)
	 * if ANYTHING doesnt go completely 100% okay
	 */
}

static int fb_open(dev_t *dev, int openflags, int otyp, cred_t *credp){
	int instance=DEVT_TO_INST(*dev);
	struct fb_softstate *softstate;

   	fb_debugprint("fb: open called\n");

	/* even if you dont USE this, it is a check to see if we need
	 * to return ENXIO
	 */
	softstate=ddi_get_soft_state(fb_glob_soft_handle, instance);
	if(softstate==NULL){
		/* This is technically NOT an error; it is expected.
		 * We return ENXIO to support "deferred attach"
		 */
		return (ENXIO);
	}
	/* return 0 on okay, or an errno otherwise */
	return 0;
}



static int fb_read(dev_t dev, struct uio* uiop, cred_t * credp){
	int instance=DEVT_TO_INST(dev);
	
   	fb_debugprint("fb: read called for instance %d\n",instance);
	/* If you are handling things directly here rather than through
	 * physio, make sure that uiop->resid is set to 0 on good read
	 */
	return (ENXIO);
}

static int fb_write(dev_t dev, struct uio *uiop, cred_t *credp){
	int instance=DEVT_TO_INST(dev);

   	fb_debugprint("fb: write called for instance %d\n",instance);
	/* If you are handling things directly here rather than through
	 * physio, make sure that uiop->resid is set to 0 on good write
	 */
   	return (ENXIO);
}

static int fb_close(dev_t dev, int openflags, int otyp, cred_t *credp){
   	fb_debugprint("fb: close called\n");
   	return (ENXIO);
}

/* For now, we hardcode values in here.
 * pretend to be 32bitsperpixel, I guess.
 * for nvidia, anyways.
 * but otherwise, SVGA is... ?
 * btw: "man fbio" for some amount of documentation
 *
 */
static int fb_ioctl(dev_t dev, int cmd, intptr_t origarg, int flags,
		    cred_t *credp, int *rvalp){
	int *arg = (int *)origarg;
	int minornum=getminor(dev);
	int instance=minornum;
	struct fb_softstate *softstate;
	struct fbtype mytype;
	struct fbgattr mygattr;

	
	softstate=ddi_get_soft_state(fb_glob_soft_handle, instance);
	if(softstate==NULL){
		printf("ioctl called, but no softstate for instance %d\n",
		        instance);
		return ENXIO;
	}
	/* could get softstate, but right now, no point */

	mytype=softstate->mytype;
	mygattr=softstate->mygattr;

   	fb_debugprint("fb: ioctl called\n");

	bzero(&mygattr,sizeof(mygattr));
	mygattr.real_type=mytype.fb_type;
	mygattr.owner=0;
	bcopy(&mytype, &mygattr.fbtype, sizeof(struct fbtype));

	mygattr.sattr.flags=0; /* or FB_ATTR_AUTOINIT...? */
	mygattr.sattr.emu_type=-1;
	/* mygattr.dev_specific[xx]=?? */

	switch(cmd){
	    case FBIOGTYPE:
		fb_debugprint(" (FBIOGTYPE)\n");
		
		ddi_copyout(&mytype, arg, sizeof(mytype), flags);
		return 0;

	    case FBIOGATTR:
		fb_debugprint(" (FBIOGATTR)\n");
		/* for some reason, sun had the original GTYPE, and
		 * then this one, that gets the "real device type".
		 * Arrg.
		 */
		ddi_copyout(&mygattr, arg, sizeof(mygattr), flags);
		return 0;

	    case FBIOSATTR:
		fb_debugprint(" (FBIOSATTR)\n");
		return (ENXIO);

	    case FBIOMONINFO:
		fb_debugprint(" (FBIOMONINFO)\n");
		return (ENXIO);


	    default:
		fb_debugprint("unknown/unsupported ioctl %d\n",cmd);
		return (ENXIO);
	}
		
	/* reminder: */
	/* ddi_copyin(arg, &deststruct, sizeof(dest), flags); */

	
}

static int fb_detach(dev_info_t *dip, ddi_detach_cmd_t cmd){
   	int instance;
	struct fb_softstate *softstate;
   	instance=ddi_get_instance(dip);

	fb_debugprint("fb: detach for instance %d\n", instance);

	softstate=ddi_get_soft_state(fb_glob_soft_handle, instance);
	if(softstate==NULL){
		cmn_err(CE_NOTE,"fb_detach failed: cannot get softstate?");
		return DDI_FAILURE;
	}

	fb_cleanup(softstate);

	ddi_remove_minor_node(dip, DEVNODE_NAME);

	return (DDI_SUCCESS);
		     
}

